@@ -1,7 +1,7 @@
 #! /usr/bin/env python
 ## vim: fileencoding=utf-8
 #
-# Copyright (c) 2007 Adeodato Simó (dato@net.com.org.es)
+# Copyright (c) 2007-2008 Adeodato Simó (dato@net.com.org.es)
 # Licensed under the terms of the MIT license.
 
 import os
@@ -30,6 +30,7 @@ class Playlist(kdeui.KListView, util.HasConfig, util.HasGUIConfig):
         util.HasGUIConfig.__init__(self)
 
         self.queue = []
+        self.visualizer_rect = None
         self.columns = Columns(self)
         self.stop_mode = StopMode.NONE
         self.random_queue = util.RandomOrderedList()
@@ -235,13 +236,17 @@ class Playlist(kdeui.KListView, util.HasConfig, util.HasGUIConfig):
 
     def slot_accept_drop(self, event, prev_item):
         if event.source() != self.viewport(): # XXX
+            # If Control is pressed, we want to append at the end:
+            if (kdecore.KApplication.kApplication().keyboardMouseState()
+                    & qt.Qt.ControlButton):
+                prev_item = self.lastItem()
             files = drag.FileListDrag.file_list(event)
             self.add_files(files, prev_item)
 
     def slot_clear(self):
         self.queue[:] = []
         self.random_queue[:] = []
-        self.tag_reader.clear_queue()
+        self.tag_reader.worker.clear_queue()
 
         if self._currently_playing not in (self.FIRST_ITEM, None):
             # We don't want the currently playing item to be deleted,
@@ -264,7 +269,7 @@ class Playlist(kdeui.KListView, util.HasConfig, util.HasGUIConfig):
 
         for item in items:
             self.takeItem(item)
-            self.tag_reader.dequeue(item)
+            self.tag_reader.worker.dequeue(item)
             self.toggle_enqueued(item, only_dequeue=True)
             try:
                 self.random_queue.remove(item)
@@ -624,7 +629,7 @@ class Playlist(kdeui.KListView, util.HasConfig, util.HasGUIConfig):
 
         if self._regex_mode == 'Always' or (regex_failed
                 and self._regex_mode == 'OnRegexFail'):
-            self.tag_reader.queue(item)
+            self.tag_reader.worker.queue(item)
 
         return item
 
@@ -662,6 +667,28 @@ class Playlist(kdeui.KListView, util.HasConfig, util.HasGUIConfig):
         else:
             return kdeui.KListView.acceptDrag(self, event)
 
+    def contentsDragMoveEvent(self, event):
+        if (not (kdecore.KApplication.kApplication().keyboardMouseState()
+                    & qt.Qt.ControlButton)
+                or not drag.FileListDrag.canDecode(event)):
+            if self.visualizer_rect is not None:
+                self.viewport().repaint(self.visualizer_rect, True)
+                self.visualizer_rect = None
+            return kdeui.KListView.contentsDragMoveEvent(self, event)
+        else:
+            try:
+                self.cleanDropVisualizer()
+                self.setDropVisualizer(False)
+                rect = self.drawDropVisualizer(None, None, self.lastChild())
+                if rect != self.visualizer_rect:
+                    self.visualizer_rect = rect
+                    brush = qt.QBrush(qt.Qt.Dense4Pattern)
+                    painter = qt.QPainter(self.viewport())
+                    painter.fillRect(self.visualizer_rect, brush)
+                return kdeui.KListView.contentsDragMoveEvent(self, event)
+            finally:
+                self.setDropVisualizer(True)
+
     def eventFilter(self, object_, event):
         # TODO Avoid so many calls to viewport(), type(), button(), ... ?
 
@@ -858,9 +885,8 @@ class PlaylistItem(kdeui.KListViewItem):
             e_height = self.height() - e_margin*2
 
             if draw_stop:
-                stop_pixmap = util.get_png('black_tiny_stop')
-                s_width = stop_pixmap.width()
-                s_height = stop_pixmap.height()
+                s_width = 8
+                s_height = 8
             else:
                 s_width = s_height = 0
 
@@ -883,7 +909,8 @@ class PlaylistItem(kdeui.KListViewItem):
 
             if draw_stop:
                 y = e_height / 2 - s_height / 2 + e_margin
-                painter.drawPixmap(x, y, stop_pixmap)
+                painter.setBrush(qt.QColor(0, 0, 0))
+                painter.drawRect(x, y, s_width, s_height)
                 x += s_width + e_margin/2
 
             if queue_pos:
